/* Copyright (C) 2004 - 2007  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
package com.db4o.foundation;


/**
 * @exclude
 */
public class Hashtable4 implements DeepClone {

	private static final float FILL = 0.5F;

	private int i_tableSize;

	private int i_mask;

	private int i_maximumSize;

	private int i_size;

	private HashtableIntEntry[] i_table;

	public Hashtable4(int a_size) {
		a_size = newSize(a_size); // legacy for .NET conversion
		i_tableSize = 1;
		while (i_tableSize < a_size) {
			i_tableSize = i_tableSize << 1;
		}
		i_mask = i_tableSize - 1;
		i_maximumSize = (int) (i_tableSize * FILL);
		i_table = new HashtableIntEntry[i_tableSize];
	}

	public Hashtable4() {
		this(1);
	}
	
	protected Hashtable4(DeepClone cloneOnlyCtor) {
	}
	
	public int size() {
		return i_size;
	}

	public Object deepClone(Object obj) {
		return deepCloneInternal(new Hashtable4((DeepClone)null), obj);
	}

	public void forEachKey(Visitor4 visitor) {
		for (int i = 0; i < i_table.length; i++) {
			HashtableIntEntry entry = i_table[i];
			while (entry != null) {
				entry.acceptKeyVisitor(visitor);
				entry = entry.i_next;
			}
		}
	}

	public void forEachKeyForIdentity(Visitor4 visitor, Object a_identity) {
		for (int i = 0; i < i_table.length; i++) {
			HashtableIntEntry entry = i_table[i];
			while (entry != null) {
				if (entry.i_object == a_identity) {
					entry.acceptKeyVisitor(visitor);
				}
				entry = entry.i_next;
			}
		}
	}

	public void forEachValue(Visitor4 visitor) {
		for (int i = 0; i < i_table.length; i++) {
			HashtableIntEntry entry = i_table[i];
			while (entry != null) {
				visitor.visit(entry.i_object);
				entry = entry.i_next;
			}
		}
	}

	public Object get(byte[] key) {
		int intKey = HashtableByteArrayEntry.hash(key);
		return getFromObjectEntry(intKey, key);
	}

	public Object get(int key) {
		HashtableIntEntry entry = i_table[key & i_mask];
		while (entry != null) {
			if (entry.i_key == key) {
				return entry.i_object;
			}
			entry = entry.i_next;
		}
		return null;
	}

	public Object get(Object key) {
		if (key == null) {
			return null;
		}
		return getFromObjectEntry(key.hashCode(), key);
	}
	
	public boolean containsKey(Object key) {
		if (null == key) {
			return false;
		}
		return null != getObjectEntry(key.hashCode(), key); 
	}

	public void put(byte[] key, Object value) {
		putEntry(new HashtableByteArrayEntry(key, value));
	}

	public void put(int key, Object value) {
		putEntry(new HashtableIntEntry(key, value));
	}

	public void put(Object key, Object value) {
		putEntry(new HashtableObjectEntry(key, value));
	}

	public Object remove(byte[] key) {
		int intKey = HashtableByteArrayEntry.hash(key);
		return removeObjectEntry(intKey, key);
	}

	public void remove(int a_key) {
		HashtableIntEntry entry = i_table[a_key & i_mask];
		HashtableIntEntry predecessor = null;
		while (entry != null) {
			if (entry.i_key == a_key) {
				removeEntry(predecessor, entry);
				return;
			}
			predecessor = entry;
			entry = entry.i_next;
		}
	}

	public void remove(Object objectKey) {
		int intKey = objectKey.hashCode();
		removeObjectEntry(intKey, objectKey);
	}

	protected Hashtable4 deepCloneInternal(Hashtable4 ret, Object obj) {
		ret.i_mask = i_mask;
		ret.i_maximumSize = i_maximumSize;
		ret.i_size = i_size;
		ret.i_tableSize = i_tableSize;
		ret.i_table = new HashtableIntEntry[i_tableSize];
		for (int i = 0; i < i_tableSize; i++) {
			if (i_table[i] != null) {
				ret.i_table[i] = (HashtableIntEntry) i_table[i].deepClone(obj);
			}
		}
		return ret;
	}

	private int entryIndex(HashtableIntEntry entry) {
		return entry.i_key & i_mask;
	}

	private HashtableIntEntry findWithSameKey(HashtableIntEntry newEntry) {
		HashtableIntEntry existing = i_table[entryIndex(newEntry)];
		while (null != existing) {
			if (existing.sameKeyAs(newEntry)) {
				return existing;
			}
			existing = existing.i_next;
		}
		return null;
	}

	private Object getFromObjectEntry(int intKey, Object objectKey) {
		final HashtableObjectEntry entry = getObjectEntry(intKey, objectKey);		
		return entry == null ? null : entry.i_object;
	}

	private HashtableObjectEntry getObjectEntry(int intKey, Object objectKey) {
		HashtableObjectEntry entry = (HashtableObjectEntry) i_table[intKey & i_mask];
		while (entry != null) {
			if (entry.i_key == intKey && entry.hasKey(objectKey)) {
				return entry;
			}
			entry = (HashtableObjectEntry) entry.i_next;
		}
		return null;
	}

	private void increaseSize() {
		i_tableSize = i_tableSize << 1;
		i_maximumSize = i_maximumSize << 1;
		i_mask = i_tableSize - 1;
		HashtableIntEntry[] temp = i_table;
		i_table = new HashtableIntEntry[i_tableSize];
		for (int i = 0; i < temp.length; i++) {
			reposition(temp[i]);
		}
	}

	private void insert(HashtableIntEntry newEntry) {
		i_size++;
		if (i_size > i_maximumSize) {
			increaseSize();
		}
		int index = entryIndex(newEntry);
		newEntry.i_next = i_table[index];
		i_table[index] = newEntry;
	}

	private final int newSize(int a_size) {
		return (int) (a_size / FILL);
	}

	private void putEntry(HashtableIntEntry newEntry) {
		HashtableIntEntry existing = findWithSameKey(newEntry);
		if (null != existing) {
			replace(existing, newEntry);
		} else {
			insert(newEntry);
		}
	}

	private void removeEntry(HashtableIntEntry predecessor, HashtableIntEntry entry) {
		if (predecessor != null) {
			predecessor.i_next = entry.i_next;
		} else {
			i_table[entryIndex(entry)] = entry.i_next;
		}
		i_size--;
	}

	private Object removeObjectEntry(int intKey, Object objectKey) {
		HashtableObjectEntry entry = (HashtableObjectEntry) i_table[intKey & i_mask];
		HashtableObjectEntry predecessor = null;
		while (entry != null) {
			if (entry.i_key == intKey && entry.hasKey(objectKey)) {
				removeEntry(predecessor, entry);
				return entry.i_object;
			}
			predecessor = entry;
			entry = (HashtableObjectEntry) entry.i_next;
		}
		return null;
	}

	private void replace(HashtableIntEntry existing, HashtableIntEntry newEntry) {
		newEntry.i_next = existing.i_next;
		HashtableIntEntry entry = i_table[entryIndex(existing)];
		if (entry == existing) {
			i_table[entryIndex(existing)] = newEntry;
		} else {
			while (entry.i_next != existing) {
				entry = entry.i_next;
			}
			entry.i_next = newEntry;
		}
	}

	private void reposition(HashtableIntEntry a_entry) {
		if (a_entry != null) {
			reposition(a_entry.i_next);
			a_entry.i_next = i_table[entryIndex(a_entry)];
			i_table[entryIndex(a_entry)] = a_entry;
		}
	}
}
